import os, re
from webob.request import BaseRequest, environ_from_url
from webob.response import Response as WebobResponse, iter_close
from mext.uri import AbsoluteUri
from mext.uri.encode import _new_quote_table, _qt_safe2, charmap_decode
from mext.uri.parse import parse_query_unicode


_on_gae = os.environ.get('SERVER_SOFTWARE', '').startswith('Google')

if _on_gae:
    from cStringIO import StringIO
    # wrapper is necessary to make f.name settable
    class temp_file(object):
        def __init__(self):
            f = StringIO()
            for name in ('read', 'readline', 'seek', 'tell', 'truncate', 'flush', 'close'):
                setattr(self, name, getattr(f, name))
    part_file = temp_file
else:
    try:
        from mext.tempf import mem_file, temp_file
        part_file = mem_file
    except ImportError:
        import tempfile
        part_file = temp_file = tempfile.TemporaryFile




__all__ = ['Request', 'Response', 'OverrideResponse', 'wapp', 'wapp_mw']



#@@ override cookies

def _set_attrs(ob, kw):
    cls = ob.__class__
    for name, value in kw.iteritems():
        if not hasattr(cls, name):
            raise TypeError("Unexpected keyword: %s=%r" % (name, value))
        setattr(ob, name, value)

_qt_path = _new_quote_table(_qt_safe2, '', '?#')
def quote_path(s):
    return str(charmap_decode(s, 'strict', _qt_path)[0])


@property
def masked_prop(self):
    raise AttributeError("This property is masked by PasteOb")

class Request(BaseRequest):
    charset = property(lambda self: 'UTF-8') # r/o
    str_GET = str_POST = str_params = _charset_cache = urlvars = masked_prop #@@str_cookies

    def __init__(self, env='/', **kw):
        if isinstance(env, basestring):
            if isinstance(env, unicode):
                env = env.encode('utf8')
            env = environ_from_url(env)
        env.setdefault('QUERY_STRING', '')
        self.environ = env
        if kw:
            _set_attrs(self, kw)

    def get_response(self, app):
        status, headers, app_iter = self.call_application(app)
        return Response(status=status, headerlist=list(headers), app_iter=app_iter)
    __rlshift__ = get_response # resp = app<<req

    @property
    def GET(self):
        e = self.environ
        qs = e['QUERY_STRING']
        if not qs:
            return {}
        cached = e.get('pasteob.GET', (None, None))
        if cached[0] == qs:
            return cached[1]
        r = parse_query_unicode(qs)
        e['pasteob.GET'] = (qs, r)
        return r

    @property
    def POST(self):
        e = self.environ
        if 'pasteob.POST' in e:
            return e['pasteob.POST']
        if self.method != 'POST':
            r = {}
        elif self.content_type == 'multipart/form-data':
            ctype = self.environ['CONTENT_TYPE']
            boundary = _rx_boundary.search(ctype).group(1).strip('"')
            r = _read_form_data(self.body_file, boundary)
        elif self.content_type in ('application/x-www-form-urlencoded', ''):
            r = parse_query_unicode(self.body)
        else:
            r = {}
        e['pasteob.POST'] = r
        return r

    @property
    def params(self):
        g, p = self.GET, self.POST
        if g and p:
            r = g.copy()
            r.update(p)
            return r
        else:
            return g or p

    @property
    def path(self):
        e = self.environ
        return quote_path(e['SCRIPT_NAME']+e['PATH_INFO'])

    @property
    def path_qs(self):
        e = self.environ
        r = quote_path(e['SCRIPT_NAME']+e['PATH_INFO'])
        qs = e['QUERY_STRING']
        if qs:
            return r + '?' + qs
        return r

    @property
    def host_url(self):
        e = self.environ
        url = e['wsgi.url_scheme'] + '://'
        host = e.get('HTTP_HOST')
        if not host:
            url += e['SERVER_NAME']
            port = e['SERVER_PORT']
        elif ':' in host:
            host, port = host.split(':', 1)
            url += host
        else:
            return url+host

        if ((port == '80' and e['wsgi.url_scheme'] == 'http')
            or (port == '443' and e['wsgi.url_scheme'] == 'https')
        ):
            return url
        else:
            return url + (':%s' % port)

    @property
    def application_url(self):
        return self.host_url + quote_path(self.environ['SCRIPT_NAME'])

    @property
    def url(self):
        return self.host_url + self.path_qs

#     @property
#     def url_no_query(self):
#         return self.host_url + self.path
#     path_url = url_no_query # webob uses this name

    uri = property(lambda self: AbsoluteUri(self.url))
    app_uri = property(lambda self: AbsoluteUri(self.application_url))
    appdir_uri = property(lambda self: AbsoluteUri(self.application_url).as_dir())
    host_uri = property(lambda self: AbsoluteUri(self.host_url))

    @property
    def referrer_uri(self):
        r = self.referrer
        return AbsoluteUri(r) if r else None

    # no disk access on GAE
    request_body_tempfile_limit = None if _on_gae else 1<<18 # 256K
    @staticmethod
    def make_tempfile():
        return temp_file()






class Response(WebobResponse):
    ctype = WebobResponse.content_type
    conditional_response = property()
    RequestClass = Request

    # when overriding __call__, override this as well
    def __lshift__(self, req):
        return self.__class__(
            status=self.status,
            headerlist=self.headerlist,
            app_iter=self.app_iter
        )

    _status = '200 OK'
    _environ = _request = None
    _body = _app_iter = _headers = None

    def __init__(self,
        body=None, status=None, headerlist=None, app_iter=None,
        ctype=None, charset=None, content_type=None,
        **kw
    ):
        if status is not None:
            self.status = status
        if ctype is None:
            ctype = content_type
        else:
            assert content_type is None

        if headerlist is not None:
            assert ctype is charset is None
            self._headerlist = headerlist
        elif ctype is None:
            assert charset is None
            self._headerlist = [('Content-Type', 'text/html; charset=UTF-8')]
            charset = 'UTF-8'
        else:
            if charset:
                ctype += '; charset=' + charset
            elif 'charset=' in ctype:
                pass
                # we can detect charset for ubody, but that's a very strange case
                #charset = ctype.rsplit('charset=', 1)[1]
            elif charset is not None:
                # pass charset='' or charset=False (anything false but not None)
                # to disable automatic utf-8 charset for text and xml responses
                pass
            elif (ctype.startswith('text/')
                or (ctype.startswith('application/') and ctype.endswith('xml'))
            ):
                # assert charset is None
                ctype += '; charset=UTF-8'
                charset = 'UTF-8'
            self._headerlist = [('Content-Type', ctype)]

        if app_iter is not None:
            assert body is None
            self._app_iter = app_iter
        else:
            body = body or ''
            if isinstance(body, unicode):
                assert charset, "Got unicode body but no charset"
                body = body.encode(charset)
            self._body = body
            assert not headerlist or ('Content-Length' not in self.headers)
            self._headerlist.append(('Content-Length', str(len(body))))
        if kw:
            _set_attrs(self, kw)

    def __call__(self, environ, start_response):
        start_response(self.status, self.headerlist)
        return self.app_iter

    def flat_app_iter(self):
        r = self._app_iter
        if r is None:
            return [self._body]
        if not isinstance(r, (list, tuple)):
            r = self._app_iter = list(r)
        return r


class OverrideResponse(BaseException):
    def __init__(self, app, comment=None):
        self.app = app
        self.comment = comment





class wapp(object):
    __slots__ = 'func'
    def __init__(self, func):
        self.func = func

    def __repr__(self):
        return '%s(%r)' % (self.__class__.__name__, self.func)

    def __get__(self, ob, type):
        try:
            get = self.func.__get__
        except:
            return self
        else:
            return self.__class__(get(ob, type))

    _trace_comment = None

    def __call__(self, env, start):
        _comment = self._trace_comment # mext.debug comment hook
        try:
            app = self.func(Request(env))
            if app is None:
                app = self.error_404
        except OverrideResponse, exc:
            app = exc.app
        return app(env, start)

    def customize(self, *args, **kw):
        def customized_func(req):
            return self.func(req, *args, **kw)
        return wapp(customized_func).__call__




class wapp_mw(object):
    __slots__ = ('mwfunc', 'next_app', 'args', 'kw')
    def __new__(cls, *args, **kw):
        if len(args) == 1 and not kw:
            mwfunc = args[0]
            factory = lambda *args, **kw: cls(mwfunc, *args, **kw)
            factory.__name__ = 'wapp_mw(%r)' % mwfunc
            return factory
        return object.__new__(cls)

    def __init__(self, mwfunc, next_app, *args, **kw):
        self.mwfunc = mwfunc
        self.next_app = next_app
        self.args = args
        self.kw = kw

    def __repr__(self):
        kwrepr = ''.join(
            map(', %r'.__mod__, self.args)
            + map(', %s=%r'.__mod__, self.kw.items())
        )
        return '%s(%r, %r%s)' % (self.__class__.__name__, self.mwfunc, self.next_app, kwrepr)

    @wapp
    def __call__(self, req):
        return self.mwfunc(req, self.next_app, *self.args, **self.kw)




_rx_boundary = re.compile(';\\s*boundary=([^;]*)')
_rx_name= re.compile(';\\s*name=([^;]*)')
_rx_fn = re.compile(';\\s*filename=([^;]*)')

#@@ use surrogate files for file data (original stream + range)

def _read_form_data(src, boundary):
    next = '--%s\r\n' % boundary
    last = '--%s--\r\n' % boundary
    sep = (next, last, '')
    r = {}
    f = part_file()
    valfn = object() # marker for regular value field
    while True:
        line = src.readline()
        if line not in sep:
            f.write(line)
            continue
        if line:
            f.seek(-2,2)
            f.truncate()
        f.seek(0)
        if getattr(f, 'name', None) is valfn:
            r[name] = f.read().decode('utf8')
        if line != next:
            return r
        headers = _read_part_headers(src)
        cdisp = headers.get('content-disposition', '')
        f = part_file()
        if not cdisp.startswith('form-data;'):
            continue
        name = _rx_name.search(cdisp)
        if not name:
            continue
        name = _decode_param(name.group(1))
        r[name] = f
        fn = _rx_fn.search(cdisp)
        if fn:
            f.name = _decode_param(fn.group(1))
        else:
            f.name = valfn


def _read_part_headers(src):
    r = {}
    while True:
        line = src.readline().strip()
        if not line:
            return r
        name, val = line.split(':', 1)
        r[name.lower()] = val.lstrip()

def _decode_param(val):
    if len(val) >= 2 and val[0] == val[-1] == '"':
        return val[1:-1].replace('\\\\', '\\').replace('\\"', '"').decode('utf8')
    return val
